/*
 * System tag database
 *
 * Copyright (C) 2003  Enrico Zini <enrico@debian.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 */

#include <debtags/Tags.h>
#include <debtags/Paths.h>
#include <debtags/Vocabulary.h>

#include <tagcoll/StdioParserInput.h>
#include <tagcoll/TextFormat.h>
#include <tagcoll/Patches.h>
#include <tagcoll/Filters.h>

#include <sys/wait.h>	// WIFEXITED WEXITSTATUS
#include <sys/types.h>	// getpwuid, stat, mkdir, getuid
#include <sys/stat.h>	// stat, mkdir
#include <pwd.h>		// getpwuid
#include <unistd.h>		// stat, getuid

#include <errno.h>

using namespace std;
using namespace Debtags;

static std::string rcdir;

static void init_rcdir(bool editable) throw (SystemException, ConsistencyCheckException)
{
	struct passwd* udata = getpwuid(getuid());
	rcdir = udata->pw_dir;
	rcdir += "/.debtags";

	//printf("Init debtags environment; rcdir is %.*s\n", PFSTR(rcdir));

	struct stat rcdir_stat;
	if (stat(rcdir.c_str(), &rcdir_stat) == -1)
	{
		//printf("Creating profile directory %.*s.\n", PFSTR(rcdir));
		if (editable && mkdir(rcdir.c_str(), 0777) == -1)
			throw SystemException(errno, "creating directory " + rcdir);
	} else {
		if (S_ISDIR(rcdir_stat.st_mode) == 0)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and is not a directory");
			else
				fprintf(stderr, "warning: %.*s already exists and is not a directory", PFSTR(rcdir));
		
		if (access(rcdir.c_str(), R_OK | W_OK | X_OK) == -1)
			if (editable)
				throw ConsistencyCheckException(rcdir + " already exists and has insufficient permissions");
			else
				fprintf(stderr, "%.*s already exists and has insufficient permissions", PFSTR(rcdir));
	}
}


namespace Debtags
{

template<class ITEM>
Tags<ITEM>::Tags(
		Converter<ITEM, std::string>& fromitem,
		Converter<std::string, ITEM>& toitem,
		bool editable)
	: TDBReadonlyDiskIndex<ITEM, Tag>(
			string(Paths::path_tagdb_indexpkg),
			string(Paths::path_tagdb_indextag),
			fromitem, fromtag, toitem, totag),
	  voc(Paths::path_vocabulary, Paths::path_vocabulary_index),
	  fromitem(fromitem),
	  toitem(toitem), totag(voc)
{
	if (rcdir.empty())
		init_rcdir(editable);

	string patchFile = rcdir + "/patch";
	if (access(patchFile.c_str(), F_OK) == 0)
	{
		StdioParserInput in(patchFile);
		PatchList<ITEM, Tag> patch = TextFormat<ITEM, Tag>::parsePatch(this->toitem, this->totag, in);
		setChanges(patch);
	}
}


template<class ITEM>
bool Tags<ITEM>::hasTagDatabase()
{
	if (access(Paths::path_tagdb, R_OK) == -1)
	{
		fprintf(stderr, "Missing tag database %s\n", Paths::path_tagdb);
		return false;
	}
	if (access(Paths::path_tagdb_indexpkg, R_OK) == -1)
	{
		fprintf(stderr, "Missing package index for tag database %s\n", Paths::path_tagdb_indexpkg);
		return false;
	}
	if (access(Paths::path_tagdb_indextag, R_OK) == -1)
	{
		fprintf(stderr, "Missing tag index for tag database %s\n", Paths::path_tagdb_indextag);
		return false;
	}
	if (access(Paths::path_vocabulary, R_OK) == -1)
	{
		fprintf(stderr, "Missing tag vocabulary %s\n", Paths::path_vocabulary);
		return false;
	}
	if (access(Paths::path_vocabulary_index, R_OK) == -1)
	{
		fprintf(stderr, "Missing index for tag vocabulary %s\n", Paths::path_vocabulary_index);
		return false;
	}
	return true;
}


template<class ITEM>
void Tags<ITEM>::savePatch()
{
	savePatch(this->getChanges());
}

template<class ITEM>
void Tags<ITEM>::savePatch(const PatchList<ITEM, Debtags::Tag>& patch)
{
	string patchFile = rcdir + "/patch";
	string backup = patchFile + "~";

	if (access(patchFile.c_str(), F_OK) == 0)
		if (rename(patchFile.c_str(), backup.c_str()) == -1)
			throw SystemException(errno, "Can't rename " + patchFile + " to " + backup);

	try {
		FILE* out = fopen(patchFile.c_str(), "w");
		if (out == 0)
			throw SystemException(errno, "Can't write to " + patchFile);
		
		TextFormat<ITEM, Tag>::outputPatch(fromitem, fromtag, patch, out);

		fclose(out);
	} catch (Exception& e) {
		if (rename(backup.c_str(), patchFile.c_str()) == -1)
			fprintf(stderr, "Warning: Impossible to restore previous backup copy: %.*s (%s)\n", PFSTR(e.desc()), e.type());
		throw;
	}
}

template<class ITEM>
void Tags<ITEM>::sendPatch()
{
	PatchList<ITEM, Tag> patch = this->getChanges();
	if (!patch.empty())
		sendPatch(this->getChanges());
}

template<class ITEM>
void Tags<ITEM>::sendPatch(const PatchList<ITEM, Debtags::Tag>& patch)
{
	static const char* cmd = "/usr/sbin/sendmail -t";
	FILE* out = popen(cmd, "w");
	if (out == 0)
		throw SystemException(errno, string("trying to run `") + cmd + "'");

	struct passwd* udata = getpwuid(getuid());

	fprintf(out,
			"To: debtags-submit@vitavonni.de\n"
			"Bcc: %s\n"
			"Subject: Tag patch\n"
			"Mime-Version: 1.0\n"
			"Content-Type: multipart/mixed; boundary=\"9amGYk9869ThD9tj\"\n"
			"Content-Disposition: inline\n"
			"X-Mailer: debtags-edit\n\n"
			"This mail contains a Debtags patch for the central archive\n\n"
			"--9amGYk9869ThD9tj\n"
			"Content-Type: text/plain; charset=utf-8\n"
			"Content-Disposition: inline\n\n"
			"-- DEBTAGS DIFF V0.1 --\n", udata->pw_name);

	TextFormat<ITEM, Tag>::outputPatch(fromitem, fromtag, patch, out);

	fprintf(out, "\n--9amGYk9869ThD9tj\n");

	int res = pclose(out);
	if (!WIFEXITED(res) || WEXITSTATUS(res) != 0)
		throw ConsistencyCheckException("sendmail returned nonzero (" + stringf::fmt(res) + "): the mail may have not been sent");
}


template<class ITEM>
void Tags<ITEM>::outputSystem(Tagcoll::Consumer<std::string, std::string>& cons)
{
	StdioParserInput in(Paths::path_tagdb);
	Converter<string, string> conv;
	TextFormat<string, string>::parse(conv, conv, in, cons);
}


template<class ITEM>
void Tags<ITEM>::outputSystem(Tagcoll::Consumer<ITEM, Tag>& cons)
{
	StdioParserInput in(Paths::path_tagdb);
	TextFormat<ITEM, Tag>::parse(this->toitem, this->totag, in, cons);
}

template<class ITEM>
void Tags<ITEM>::outputPatched(Tagcoll::Consumer<std::string, std::string>& cons)
{
	const string patchFile = rcdir + "/patch";
	if (access(patchFile.c_str(), F_OK) == 0)
	{
		StdioParserInput inpatch(patchFile);
		Converter<string, string> conv;
		PatchList<string, string> patch =
			TextFormat<string, string>::parsePatch(conv, conv, inpatch);

		patch.setConsumer(cons);

		StdioParserInput in(Paths::path_tagdb);
		TextFormat<string, string>::parse(conv, conv, in, patch);
	} else
		outputSystem(cons);
}

template<class ITEM>
void Tags<ITEM>::outputPatched(Tagcoll::Consumer<ITEM, Debtags::Tag>& cons)
{
	string patchFile = rcdir + "/patch";
	if (access(patchFile.c_str(), F_OK) == 0)
	{
		StdioParserInput inpatch(patchFile);
		PatchList<ITEM, Tag> patch = TextFormat<ITEM, Tag>::parsePatch(toitem, totag, inpatch);

		patch.setConsumer(cons);

		StdioParserInput in(Paths::path_tagdb);
		TextFormat<ITEM, Tag>::parse(toitem, totag, in, patch);
	} else
		outputSystem(cons);
}

}

#ifndef INSTANTIATING_TEMPLATES
#include <string>

namespace Debtags {
	template class Tags<std::string>;
}
#endif


#ifdef COMPILE_TESTSUITE

#include <tests/test-utils.h>

namespace tut {

struct debtags_tags_shar {
};
TESTGRP(debtags_tags);

template<> template<>
void to::test<1>()
{
	ensure(false);
}

}

#endif

// vim:set ts=4 sw=4:
