/*
 * Tag vocabulary access
 *
 * Copyright (C) 2003,2004,2005  Enrico Zini <enrico@debian.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <debtags/Vocabulary.h>
#include <debtags/TagImpl.h>
#include <debtags/Tag.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#include <errno.h>
#include <assert.h>

#include <set>

using namespace std;
using namespace Debtags;
using namespace Tagcoll;

struct ItemPos
{
	off_t off;
	unsigned int len;
};

static bool get_itempos_tdb(const TDBFile& db, const string& key, off_t* off, unsigned int* len) throw ()
{
	TDB_DATA k, v;

	k.dptr  = const_cast<char*>(key.data());
	k.dsize = key.size();

	v = tdb_fetch(const_cast<TDB_CONTEXT*>(db.context()), k);
	if (!v.dptr || v.dsize != sizeof(ItemPos))
	{
		*off = 0;
		*len = 0;
		return false;
	}

	// Deserialize
	ItemPos* ip = (ItemPos*)v.dptr;
	*off = ip->off;
	*len = ip->len;
	free(v.dptr);
	return true;
}

Vocabulary::Vocabulary(const std::string& filename, const std::string& indexname) throw (SystemException)
	: filename(filename), db(indexname), vocdata(0)
{
	// Open and mmap the file
	fd = open(filename.c_str(), O_RDONLY);
	if (fd == -1)
		throw SystemException(errno, "opening vocabulary file " + filename);

	// Get the file length
	vocdata_len = lseek(fd, 0, SEEK_END);
	if (vocdata_len == (off_t)-1)
	{
		close(fd);
		throw SystemException(errno, "getting length of vocabulary file " + filename);
	}
	// mmap the file
	if ((vocdata = (char*)mmap(0, vocdata_len, PROT_READ, MAP_PRIVATE, fd, 0)) == MAP_FAILED)
	{
		close(fd);
		throw SystemException(errno, "mapping vocabulary file " + filename + " to memory");
	}

	// Open the index
	try {
		db.open(0, O_RDONLY);
	} catch (SystemException& e) {
		munmap(vocdata, vocdata_len);
		close(fd);
		throw;
	}
}

Vocabulary::~Vocabulary() throw ()
{
	munmap(vocdata, vocdata_len);
	close(fd);
}

bool Vocabulary::hasFacet(const std::string& name) const throw ()
{
	return db.has("F"+name);
}

bool Vocabulary::hasTag(const std::string& fullname) const throw ()
{
	return db.has("T"+fullname);
}

Facet Vocabulary::getFacet(const std::string& name) const throw ()
{
	off_t ofs;
	unsigned int len;
	
	if (!get_itempos_tdb(db, "F"+name, &ofs, &len))
		return Facet();
	
	return Facet(new FacetImpl(*this, vocdata + ofs, len));
}

Tag Vocabulary::getTag(const std::string& fullname) const throw ()
{
	off_t ofs;
	unsigned int len;
	
	unsigned int p = fullname.find("::");
	Facet facet = (p == string::npos) ? getFacet("legacy") : getFacet(fullname.substr(0, p));

	/*
	if (p != string::npos)
		assert(facet.name() == fullname.substr(0, p));
	*/

	if (!facet)
		return Tag();

	if (!get_itempos_tdb(db, "T"+fullname, &ofs, &len))
		return Tag();

	return Tag(new TagImpl(facet.impl, vocdata + ofs, len));
}

OpSet<Facet> Vocabulary::getFacets() const throw ()
{
	OpSet<std::string> facetNames = db.getStringSet("f");
	OpSet<Facet> res;

	for (OpSet<string>::const_iterator i = facetNames.begin();
			i != facetNames.end(); i++)
		res += getFacet(*i);

	return res;
}

OpSet<Tag> Vocabulary::getTags() const throw ()
{
	OpSet<Facet> fs = getFacets();
	OpSet<Tag> res;
	for (OpSet<Facet>::const_iterator i = fs.begin(); i != fs.end(); i++)
		res += i->tags();
	return res;
}

OpSet<Tag> Vocabulary::getTags(const std::string& facetName) const throw ()
{
	OpSet<std::string> tagNames = db.getStringSet("t" + facetName);
	OpSet<Tag> res;

	for (set<string>::const_iterator i = tagNames.begin();
			i != tagNames.end(); i++)
		res += getTag(facetName + "::" + *i);

	return res;
}


#ifdef COMPILE_TESTSUITE

#include <tests/test-utils.h>

namespace tut {

struct debtags_vocabulary_shar {
};
TESTGRP(debtags_vocabulary);

template<> template<>
void to::test<1>()
{
	ensure(false);
}

}

#endif

// vim:set ts=4 sw=4:
