#pragma implementation

#include "config.h"

#include <debtags/Environment.h>
#include <debtags/Paths.h>
#include <debtags/Consumer.h>
#include <debtags/PackageSet.h>
#include <debtags/MaintainerSet.h>
#include <debtags/Vocabulary.h>
#include <debtags/PackageDB.h>
#include <debtags/TagDB.h>
#include <debtags/Consumer.h>
#include <debtags/Tag.h>
#include <debtags/TagSet.h>

#include <tagcoll/StringParserInput.h>
#include <tagcoll/StdioParserInput.h>

#include <debtags/Update.h>

#include <apt-pkg/tagfile.h>
#include <apt-pkg/error.h>

#include <errno.h>

using namespace std;
using namespace Debtags;


Environment* Environment::instance = 0;


namespace Debtags
{
class StandardEnvironment : public Debtags::Environment
{
protected:
	// If true, save state to ~/.debtags
	bool editable;
	// If true, remove all tags that don't have a facet
	bool facet_only;

	// State directory
	std::string rcdir;
	
	// APT Package database
	PackageDB _packagedb;

	// Debtags vocabulary
	Vocabulary _vocab;

	// Serializer
	Tagcoll::Serializer<Package, Tag> _serializer;

	// Debtags database
	TagDB<Debtags::Package> _tagDB;

#if 0
	// Packages repository
	mutable PackageSet _packages;
	// Maintainers repository
	mutable MaintainerSet _maintainers;

	mutable bool availableLoaded;
	void loadAvailable(PackageConsumer* cons = 0) const throw ();

	void fillPackageData(Package& name) const throw ();

	Package sectionToPackage(pkgTagSection& section) const throw ();
#endif

public:
	StandardEnvironment(bool editable, bool facet_only) throw (ParserException, SystemException, ConsistencyCheckException);
	virtual ~StandardEnvironment() throw () {}
	
	// Check if the tag database has been created (i.e. if debtags update has been run)
	bool hasTagDatabase() throw ();

	// Return the main package search class
#if 0
	const PackageSet& packages() const throw ()
	{
		if (!availableLoaded)
			loadAvailable();
		return _packages;
	}

	// Return the main package search class
	const MaintainerSet& maintainers() const throw ()
	{
		if (!availableLoaded)
			loadAvailable();
		return _maintainers;
	}
#endif

	// Return the APT Package Database access class
	const PackageDB& packageDB() const throw () { return _packagedb; }

	// Return the Debtags serializer
	Serializer<Package, Tag>& serializer() throw () { return _serializer; }

	// Return the Debtags Database access class
	const TagDB<Debtags::Package>& tagDB() const throw ()
	{
		return _tagDB;
	}

	// Return the Debtags Database access class
	TagDB<Debtags::Package>& tagDB() throw ()
	{
		return _tagDB;
	}

	// Return the Debtags Vocabulary access class
	const Vocabulary& vocabulary() const throw () { return _vocab; }


	// Get the Package object given the package name
	Package getPackage(const std::string& name) const throw ();

#if 0
	// Get the Maintainer object given the package name
	Maintainer getMaintainer(const std::string& name) const throw ();
#endif


	// Output the whole package list to `cons'
	void output(PackageConsumer& cons) const throw ();

	// Output the filtered package list to `cons'
	void output(PackageConsumer& cons, PackageMatcher& matcher) const throw ();

	// Output the whole package list to `cons'
	void output(Tagcoll::TagcollConsumer<Package, Tag>& cons) const throw ();

	// Output the filtered package list to `cons'
	void output(Tagcoll::TagcollConsumer<Package, Tag>& cons, PackageMatcher& matcher) const throw ();


	// Output list of packages related to `pivot' to `cons'
	void outputRelated(PackageConsumer& cons, const Package& pivot, int distance = 1) const throw ();
	// Output list of packages related to the specific tagset to `cons'
	void outputRelated(PackageConsumer& cons, const TagSet& ts, int distance = 1) const throw ();

#if 0
	// Update the Debtags database
	void updateDebtagsDatabase() throw (Exception);

	// Update the Debtags database, tracking status updates
	virtual void updateDebtagsDatabase(pkgAcquireStatus* status) throw (Exception);
#endif

	friend class AvailableReader;
};

#if 0
class PackageFiller : public PkgRecordConsumer
{
protected:
	PackageSet& _packages;
	MaintainerSet& _maintainers;
	PackageConsumer* _cons;

public:
	PackageFiller(PackageSet& packages, MaintainerSet& maintainers, PackageConsumer* cons = 0) throw () :
		_packages(packages), _maintainers(maintainers), _cons(cons) {}

	virtual void operator()(pkgRecords::Parser& parser);
};
#endif


};

void Environment::init(bool editable, bool facet_only) throw (ParserException, SystemException, ConsistencyCheckException)
{
	if (instance == 0)
		instance = new StandardEnvironment(editable, facet_only);
}


StandardEnvironment::StandardEnvironment(bool editable, bool facet_only) throw (ParserException, SystemException, ConsistencyCheckException) :
		facet_only(facet_only),
		_vocab(Paths::path_vocabulary, Paths::path_vocabulary_index),
		_serializer(_vocab),
		_tagDB(_vocab, _serializer, editable)//,
		//availableLoaded(false)
{
}

bool StandardEnvironment::hasTagDatabase() throw ()
{
	return TagDB<Debtags::Package>::hasTagDatabase();
}

#if 0
void StandardEnvironment::updateDebtagsDatabase() throw (Exception)
{
	DebtagsUpdate::updateDebtagsDatabase();
}

void StandardEnvironment::updateDebtagsDatabase(pkgAcquireStatus* status) throw (Exception)
{
	DebtagsUpdate::updateDebtagsDatabase(status);
}
#endif

#if 0
void PackageFiller::operator()(pkgRecords::Parser& parser)
{
	string name = parser.Name();

	// Create a new package if it doesn't exist yet
	Package pkg = _packages.find(name);
	if (!pkg)
	{
		pkg = Package(name);
		_packages += pkg;
	}

	string val = parser.Maintainer();
	//fprintf(stderr, "sectionToPackage: maint: %.*s\n", PFSTR(val));

	int mend = val.rfind('>');
	int mbeg = val.rfind('<', mend);
	string mail = val.substr(mbeg + 1, mend-mbeg-1);

	// Create a new maintainer if it doesn't exist yet
	Maintainer maint = _maintainers.find(mail);
	if (!maint)
	{
		maint = Maintainer(mail);
		_maintainers += maint;
	}

	// Update maintainer informations
	maint.fullname(val);
	maint.addPackage(pkg);
	pkg.maint(maint);

	pkg.sdesc(parser.ShortDesc());
	pkg.ldesc(parser.LongDesc());

	const char *start;
	const char *end;
	parser.GetRec(start,end);
	pkg.fulldata(string(start, end-start));

	if (_cons)
		_cons->consume(pkg);
}

Package StandardEnvironment::sectionToPackage(pkgTagSection& section) const throw ()
{
	string name = section.FindS("Package");

	// Create a new package if it doesn't exist yet
	Package pkg = _packages.find(name);
	if (!pkg)
	{
		pkg = Package(name);
		_packages += pkg;
	}

	string val = section.FindS("Maintainer");
	//fprintf(stderr, "sectionToPackage: maint: %.*s\n", PFSTR(val));

	int mend = val.rfind('>');
	int mbeg = val.rfind('<', mend);
	string mail = val.substr(mbeg + 1, mend-mbeg-1);

	// Create a new maintainer if it doesn't exist yet
	Maintainer maint = _maintainers.find(mail);
	if (!maint)
	{
		maint = Maintainer(mail);
		_maintainers += maint;
	}

	// Update maintainer informations
	maint.fullname(val);
	maint.addPackage(pkg);
	pkg.maint(maint);

	val = section.FindS("Description");

	// Update package informations
	int send = val.find('\n');
	pkg.sdesc(val.substr(0, send));
	pkg.ldesc(val.substr(send + 1));

	// TODO: add the other fields

	if (_error->PendingError() == true)
	{
		string err;
		string errs;
		bool first = true;
		while (_GetErrorObj()->PendingError())
		{
			_GetErrorObj()->PopMessage(err);
			if (first)
			{
				errs += err;
				first = false;
			}
			else
				errs += " - " + err;
		}
		const char* start;
		const char* stop;
		section.GetSection(start, stop);
		
		fprintf(stderr, "Parsing record gave errors:\n%.s\nRecord was:\n%.*s\n",
				PFSTR(errs), start, stop-start);
	}

	return pkg;
}

void StandardEnvironment::loadAvailable(PackageConsumer* cons) const throw ()
{
	try {
		PackageFiller filler(_packages, _maintainers, cons);
		_packagedb.iterateAll(filler);
	} catch (Exception& e) {
		fprintf(stderr, "%s: %.*s.  Aborting loading of available file\n", e.type(), PFSTR(e.desc()));
	}
	availableLoaded = true;
}

void StandardEnvironment::fillPackageData(Package& pkg) const throw ()
{
	// Fetch the package info
	pkgTagSection section;
	try {
		_packagedb.getPackageRecord(pkg.name(), section);
		sectionToPackage(section);
	} catch (PackageDBException& e) {
		//fprintf(stderr, "%s: %.*s. Ignoring package %.*s.\n", e.type(), PFSTR(e.desc()), PFSTR(pkg.name()));

		//if (pkg.sdesc().empty())
			pkg.sdesc("APT database lacks package description for " + pkg.name());
		//if (pkg.ldesc().empty())
			pkg.ldesc(" This may happen if the available file, the APT database or\n"
					  " the Debtags database are a bit out of sync, or in case you\n"
					  " have local packages installed which are not present in any\n"
					  " APT source.\n\n");
		//if (!pkg.maint())
		//{
			string mail = "error";
			string fullname = "Maintainer Data Unavailable <error>";

			// Create a new maintainer if it doesn't exist yet
			Maintainer maint = _maintainers.find(mail);
			if (!maint)
			{
				maint = Maintainer(mail);
				_maintainers += maint;
			}

			// Update maintainer informations
			maint.fullname(fullname);
			maint.addPackage(pkg);
			pkg.maint(maint);
		//}
			
	}
}
#endif

Package StandardEnvironment::getPackage(const std::string& name) const throw ()
{
#if 0
	// First try searching the pool
	Package p = _packages.find(name);
	if (p)
		return p;

	// Else try to see if the package exists
	if (!_packagedb.hasPackage(name))
		return Package();

	// Create the package in the repository
	p = Package(name);
	_packages += p;

	// Load package info from the APT database
	//return fillPackageData(name);

	return p;
#endif
	return _packagedb.getPackage(name);
}

#if 0
Maintainer StandardEnvironment::getMaintainer(const std::string& name) const throw ()
{
	if (!availableLoaded)
		loadAvailable();
	
	return _maintainers.find(name);
}
#endif

class PackageConsumerToTagcollConsumer : public PackageConsumer
{
protected:
	TagcollConsumer<Package, Tag>& cons;
public:
	PackageConsumerToTagcollConsumer(TagcollConsumer<Package, Tag>& cons) : cons(cons) {}

	virtual void consume(const Package& pkg) throw ()
	{
		TagSet tags = pkg.tags();
		if (tags.empty())
			cons.consume(pkg);
		else
			cons.consume(pkg, tags);
	}
};

void StandardEnvironment::output(PackageConsumer& cons) const throw ()
{
	_packagedb.iterateAll(cons);
#if 0
	if (availableLoaded)
	{
		_packages.output(cons);
	} else {
		loadAvailable(&cons);
	}
#endif
}

void StandardEnvironment::output(PackageConsumer& cons, PackageMatcher& matcher) const throw ()
{
	if (matcher.hasOnlyDebtags())
		PackageSet(_tagDB.getItems(matcher.debtagsTagset())).output(cons);
	else if (matcher.hasDebtags())
		PackageSet(_tagDB.getItems(matcher.debtagsTagset())).output(cons, matcher);
	else
	{
		PackageMatcherFilter filter(matcher, cons);
		_packagedb.iterateAll(filter);
	}

#if 0
	if (matcher.hasNonDebtags() && !availableLoaded)
	{
		PackageMatcherFilter filter(matcher, cons);
		loadAvailable(&filter);
	}
	else
	{
		if (matcher.hasOnlyMaintainer() && matcher.isMaintainerSignificant())
		{
			fprintf(stderr, "Strategy0\n");
			for (MaintainerSet::const_iterator i = _maintainers.begin();
					i != _maintainers.end(); i++)
				if (matcher.matchMaintainer(i->fullname()))
					i->packages().output(cons);
		}
		else if (matcher.isMaintainerSignificant())
		{
			fprintf(stderr, "Strategy1\n");
			for (MaintainerSet::const_iterator i = _maintainers.begin();
					i != _maintainers.end(); i++)
				if (matcher.matchMaintainer(i->fullname()))
					i->packages().output(cons, matcher);
		}
		else if (matcher.hasOnlyName())
		{
			fprintf(stderr, "Strategy2\n");
			for (PackageSet::const_iterator i = _packages.begin();
					i != _packages.end(); i++)
				if (matcher.matchName(i->name()))
					cons.consume(*i);
		}
		else if (matcher.hasOnlyDebtags())
		{
			fprintf(stderr, "Strategy3\n");
			PackageSet(_tagDB.getItems(matcher.debtagsTagset())).output(cons);
		}
		else if (matcher.hasDebtags())
		{
			fprintf(stderr, "Strategy4\n");
			PackageSet(_tagDB.getItems(matcher.debtagsTagset())).output(cons, matcher);
		}
		else
		{
			// If no special case applies, perform a generic search
			fprintf(stderr, "Strategy5\n");
			_packages.output(cons, matcher);
		}
	}
#endif
}

void StandardEnvironment::output(Tagcoll::TagcollConsumer<Package, Tag>& cons) const throw ()
{
	PackageConsumerToTagcollConsumer conv(cons);
	output(conv);
}

void StandardEnvironment::output(Tagcoll::TagcollConsumer<Package, Tag>& cons, PackageMatcher& matcher) const throw ()
{
	PackageConsumerToTagcollConsumer conv(cons);
	output(conv, matcher);
}


void StandardEnvironment::outputRelated(PackageConsumer& cons, const Package& pivot, int distance) const throw ()
{
	PackageSet(_tagDB.getRelatedItems(pivot, distance)).output(cons);
}

void StandardEnvironment::outputRelated(PackageConsumer& cons, const TagSet& ts, int distance) const throw ()
{
	PackageSet(_tagDB.getRelatedItems(ts, distance)).output(cons);
}

// vim:set ts=4 sw=4:
