#include <debtags/Tags.h>
#include <tagcoll/InputMerger.h>

#include <set>

using namespace std;
using namespace Debtags;
using namespace Tagcoll;

static string tagsetToString(const OpSet<Tag>& ts)
{
	string res;
	for (OpSet<Tag>::const_iterator i = ts.begin(); i != ts.end(); i++)
		if (i == ts.begin())
			res += i->fullname();
		else
			res += ", " + i->fullname();
	return res;
}

static string packagesetToString(const OpSet<string>& ts)
{
	string res;
	for (OpSet<string>::const_iterator i = ts.begin(); i != ts.end(); i++)
		if (i == ts.begin())
			res += *i;
		else
			res += ", " + *i;
	return res;
}

class PackageGatherer : public Consumer<string, Tag>
{
protected:
	OpSet<string> coll;

	virtual void consumeItemUntagged(const string& item)
	{
		coll += item;
	}
	virtual void consumeItem(const string& item, const OpSet<Tag>& tags)
	{
		coll += item;
	}
public:
	virtual ~PackageGatherer() {}
	OpSet<string>& getColl() { return coll; }
};


class CollectionComparer : public Consumer<string, Tag>
{
protected:
	const Collection<string, Tag>& first;
	OpSet<string> firstItems;
	int errors;
	
	virtual void consumeItemUntagged(const string& item)
	{
		if (!firstItems.contains(item))
		{
			fprintf(stderr, "Item %.*s is only on the second collection\n", PFSTR(item));
			errors++;
		}
		else if (!first.getTags(item).empty())
		{
			fprintf(stderr, "Item %.*s has tags %.*s in the first collection but it's untagged in the first one\n",
					PFSTR(item), PFSTR(tagsetToString(first.getTags(item))));
			errors++;
		}

		firstItems -= item;
	}
	virtual void consumeItem(const string& item, const OpSet<Tag>& tags)
	{
		if (!firstItems.contains(item))
		{
			fprintf(stderr, "Item %.*s is only on the second collection\n",
					PFSTR(item));
			errors++;
		}
		else if (first.getTags(item) != tags)
		{
			fprintf(stderr, "Item %.*s has different tag sets: %.*s in the first collection and %.*s in the second one\n",
					PFSTR(item),
					PFSTR(tagsetToString(first.getTags(item))),
					PFSTR(tagsetToString(tags)));
			errors++;
		}

		firstItems -= item;
	}

public:
	CollectionComparer(const Collection<string, Tag>& first) throw () : first(first), errors(0)
	{
		PackageGatherer pg;
		first.output(pg);
		firstItems = pg.getColl();
	}
	virtual ~CollectionComparer() throw () {}

	int testResults() throw ()
	{
		if (!firstItems.empty())
		{ 
			fprintf(stderr, "%d items were only on the first collection: %.*s\n",
					firstItems.size(),
					PFSTR(packagesetToString(firstItems)));
			errors++;
		}

		if (errors)
			fprintf(stderr, "%d problems reported\n", errors);

		return errors;
	}
};


int main (int argc, char *argv[])
{
	try {
		// Install the handler for unexpected exceptions
		InstallUnexpected installUnexpected;

		//Debtags::Environment::init(false);
		Converter<string, string> conv;
		Tags<string> debtags(conv, conv);

		int result = 0;
		
		// Test set operations on items of type Package
		OpSet<string> test;
		test.insert("apt");
		test.insert("apt");
		test.insert("debtags");
		test.insert("debtags");
		if (test.size() != 2)
		{
			fprintf(stderr, "%d items (I want 2)\n", test.size());
			result = 1;
		}
		test.erase("debtags");
		if (test.size() != 1)
		{
			fprintf(stderr, "%d items (I want 1)\n", test.size());
			result = 1;
		}

		// Compare two methods of getting tags
		InputMerger<string, Tag> second;
		debtags.outputPatched(second);

		CollectionComparer comparer(debtags);
		second.output(comparer);
		return comparer.testResults();
		
/*
		PackageGatherer pg;
		TagDB::outputPatched(pg);
		for (PackageSet::const_iterator i = pg.getColl().begin(); i != pg.getColl().end(); i++)
		{
			TagSet tsfirst = i->tags();
			OpSet<string> tssecond = second.getTagsetForItem(i->name());
			
			TagSet::const_iterator fi = tsfirst.begin();
			OpSet<string>::const_iterator si = tssecond.begin();
			for ( ; fi != tsfirst.end() && si != tssecond.end(); fi++, si++)
			{
				if (fi->fullname() != *si)
				{
					fprintf(stderr, "%.*s: tag differ [debtags]%.*s != [tagcoll]%.*s\n",
							PFSTR(i->name()), PFSTR(fi->fullname()), PFSTR(*si));
					result = 1;
				}
			}
			for (; fi != tsfirst.end(); fi++)
			{
				fprintf(stderr, "%.*s: tag only in debtags: %.*s\n",
						PFSTR(i->name()), PFSTR(fi->fullname()));
				result = 1;
			}
			for (; si != tssecond.end(); si++)
			{
				fprintf(stderr, "%.*s: tag only in tagcoll: %.*s\n",
						PFSTR(i->name()), PFSTR(*si));
				result = 1;
			}
		}
		return result;
*/
	} catch (Exception& e) {
		fprintf(stderr, "%s: %.*s\n", e.type(), PFSTR(e.desc()));
	}
}

// vim:set ts=4 sw=4:
